﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Encodings.BinaryVectorEncoding;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Encodings.RealVectorEncoding;

namespace HeuristicLab.Problems.Programmable {
  public sealed class Individual : IDeepCloneable {
    private readonly Dictionary<string, BinaryVector> binaryVectors;
    private readonly Dictionary<string, IntegerVector> integerVectors;
    private readonly Dictionary<string, RealVector> realVectors;
    private readonly Dictionary<string, Permutation> permutationVectors;

    public Individual(IEnumerable<KeyValuePair<string, BinaryVector>> binaryVectors = null,
      IEnumerable<KeyValuePair<string, IntegerVector>> integerVectors = null,
      IEnumerable<KeyValuePair<string, RealVector>> realVectors = null,
      IEnumerable<KeyValuePair<string, Permutation>> permutations = null) {
      // prevent unnecessary allocation of dictionaries for unused vectors
      if (binaryVectors != null) this.binaryVectors = binaryVectors.ToDictionary(x => x.Key, x => x.Value);
      if (integerVectors != null) this.integerVectors = integerVectors.ToDictionary(x => x.Key, x => x.Value);
      if (realVectors != null) this.realVectors = realVectors.ToDictionary(x => x.Key, x => x.Value);
      if (permutations != null) this.permutationVectors = permutations.ToDictionary(x => x.Key, x => x.Value);
    }
    private Individual(Individual original, Cloner cloner) {
      cloner.RegisterClonedObject(original, this);
      if (original.binaryVectors != null) {
        binaryVectors = new Dictionary<string, BinaryVector>(original.binaryVectors.Comparer);
        foreach (var param in original.binaryVectors)
          binaryVectors[param.Key] = cloner.Clone(param.Value);
      }
      if (original.integerVectors != null) {
        integerVectors = new Dictionary<string, IntegerVector>(original.integerVectors.Comparer);
        foreach (var param in original.integerVectors)
          integerVectors[param.Key] = cloner.Clone(param.Value);
      }
      if (original.realVectors != null) {
        realVectors = new Dictionary<string, RealVector>(original.realVectors.Comparer);
        foreach (var param in original.realVectors)
          realVectors[param.Key] = cloner.Clone(param.Value);
      }
      if (original.permutationVectors != null) {
        permutationVectors = new Dictionary<string, Permutation>(original.permutationVectors.Comparer);
        foreach (var param in original.permutationVectors)
          permutationVectors[param.Key] = cloner.Clone(param.Value);
      }
    }

    public object Clone() {
      return Clone(new Cloner());
    }

    public IDeepCloneable Clone(Cloner cloner) {
      return new Individual(this, cloner);
    }

    public BinaryVector BinaryVector(string name) {
      return binaryVectors[name];
    }

    public bool BinaryValue(string name, int index) {
      return binaryVectors[name][index];
    }

    public IEnumerable<string> BinaryNames {
      get { return binaryVectors != null ? binaryVectors.Keys : Enumerable.Empty<string>(); }
    }

    public IntegerVector IntegerVector(string name) {
      return integerVectors[name];
    }

    public int IntegerValue(string name, int index) {
      return integerVectors[name][index];
    }

    public IEnumerable<string> IntegerNames {
      get { return integerVectors != null ? integerVectors.Keys : Enumerable.Empty<string>(); }
    }

    public RealVector RealVector(string name) {
      return realVectors[name];
    }

    public double RealValue(string name, int index) {
      return realVectors[name][index];
    }

    public IEnumerable<string> RealNames {
      get { return realVectors != null ? realVectors.Keys : Enumerable.Empty<string>(); }
    }

    public Permutation Permutation(string name) {
      return permutationVectors[name];
    }

    public IEnumerable<string> PermutationNames {
      get { return permutationVectors != null ? permutationVectors.Keys : Enumerable.Empty<string>(); }
    }
  }
}
